"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from typing import List, Set

from loguru import logger
from prettytable import PrettyTable

from stellanow_cli.core.helpers import ProcessedFile, SkippedFile


def log_processed_and_skipped_result(
    processed_files: List[ProcessedFile],
    skipped_files: List[SkippedFile],
    events_not_found: Set[str],
):
    if processed_files:
        logger.info("\n==============================\n      PROCESSED FILES\n==============================\n")

        table = PrettyTable(["File"])

        for processed_file in processed_files:
            table.add_row(list(processed_file))

        logger.info(table)

    if skipped_files:
        logger.info("\n==============================\n      SKIPPED FILES\n==============================\n")

        table = PrettyTable(["File", "Skipping Reason"])

        for skipped_file in skipped_files:
            table.add_row(list(skipped_file))

        for event in events_not_found:
            table.add_row([event, "Missing Event Configuration"])

        logger.info(table)

        logger.info("\nSkipped Reason - Explanation:\n")
        logger.info(
            "- File Already Exist - Existing classes can't be overridden. Use --force to override this protection."
        )
        logger.info("- Missing Event Configuration - Check if the specified event exists in the Operators Console.")
        logger.info("- No Entity Associated With Event - Event exists, but it is not attached to any entity type.")


def log_summary(skipped_files: List[SkippedFile]) -> None:
    if skipped_files:
        logger.info("\n==============================\n         SUMMARY\n==============================\n")

        table = PrettyTable(["File", "Reason for not comparing"])

        # Populate the table with data from your SkippedFile instances
        for skipped_file in skipped_files:
            table.add_row([skipped_file.filename, skipped_file.reason])

        logger.info(table)

        logger.info("\nSkipped Reason - Explanation:\n")
        logger.info("- Not Auto-Generated - It looks like the file was not generated by this CLI.")
        logger.info("- Event ID Not Found - The ID of the event is missing in the header comment.")



"""
Copyright (C) 2022-2023 Stella Technologies (UK) Limited.

This software is the proprietary information of Stella Technologies (UK) Limited.
Use, reproduction, or redistribution of this software is strictly prohibited without
the express written permission of Stella Technologies (UK) Limited.
All rights reserved.
"""

import json
import os
import tempfile
import typing as t
import uuid
import zipfile
from dataclasses import asdict, dataclass, field
from datetime import datetime
from time import sleep

import click
import paho.mqtt.client as mqtt
from loguru import logger

from stellanow_cli.core.validators import url_validator, uuid_validator, zip_file_validator


@dataclass
class EntityTypeIds:
    entityTypeDefinitionId: str
    entityId: str


@dataclass
class Metadata:
    eventTypeDefinitionId: str
    entityTypeIds: t.List[EntityTypeIds]
    messageId: str = field(default_factory=lambda: str(uuid.uuid4()))
    messageOriginDateUTC: str = field(default_factory=lambda: datetime.utcnow().isoformat() + "Z")


@dataclass
class Value:
    metadata: Metadata
    payload: str


@dataclass
class Key:
    organizationId: str
    projectId: str


@dataclass
class MsgWrapper:
    key: Key
    value: Value


def create_message(org_id, project_id, event_type, entity_type, entity_id, payload) -> MsgWrapper:
    entity_type_ids = EntityTypeIds(entityTypeDefinitionId=entity_type, entityId=entity_id)
    metadata = Metadata(eventTypeDefinitionId=event_type, entityTypeIds=[entity_type_ids])
    value = Value(metadata=metadata, payload=payload)
    key = Key(organizationId=org_id, projectId=project_id)

    return MsgWrapper(key=key, value=value)


def on_connect(client, userdata, flags, reason_code, properties):  # noqa
    if reason_code == 0:
        logger.info("Connected to MQTT Broker!")
    else:
        logger.error(f"Failed to connect, return code {reason_code}\n")


def on_publish(client, userdata, mid, reason_code, properties):  # noqa
    logger.info(f"Message published: {mid}")


@click.command()
@click.option(
    "--mqtt_username",
    required=True,
    prompt=True,
    help="Username to authenticate with MQTT broker.",
)
@click.option(
    "--mqtt_password",
    required=True,
    prompt=True,
    hide_input=True,
    help="Password to authenticate with MQTT broker.",
)
@click.option(
    "--mqtt_broker",
    required=True,
    prompt=True,
    callback=url_validator,
    help="MQTT broker to connect to.",
)
@click.option(
    "--mqtt_port",
    default=8083,
    help="Port for MQTT broker connection, default is 8883 for SSL.",
)
@click.option(
    "--input_file",
    required=True,
    prompt=True,
    type=click.Path(exists=True),
    callback=zip_file_validator,
    help="File containing game state data to send.",
)
@click.option(
    "--org_id",
    required=True,
    prompt=True,
    callback=uuid_validator,
    help="Stella Now Organization I D.",
)
@click.option(
    "--project_id",
    required=True,
    prompt=True,
    callback=uuid_validator,
    help="Stella Now Project ID.",
)
@click.option(
    "--event_type",
    required=True,
    prompt=True,
    help="Event type definition ID for the game state data.",
)
@click.option(
    "--entity_type",
    required=True,
    prompt=True,
    help="Entity type definition ID for the game state data.",
)
@click.option("--entity_id", required=True, prompt=True, help="Entity ID for the game state data.")
@click.option(
    "--infinite",
    default=False,
    is_flag=True,
    help="If set, the game state data will be sent continuously until the process is interrupted.",
)
def simulate_game_match(
    mqtt_username: str,
    mqtt_password: str,
    mqtt_broker: str,
    mqtt_port: int,
    input_file: str,
    org_id: str,
    project_id: str,
    event_type: str,
    entity_type: str,
    entity_id: str,
    infinite: bool,
    *args,
    **kwargs,
) -> None:
    """Simulate a full match of a game by sending game state data to a MQTT broker."""
    ...

    topic = f"in/{org_id}"

    # Initialize the MQTT Client
    client = mqtt.Client(
        mqtt.CallbackAPIVersion.VERSION2,
        transport="websockets",
        protocol=mqtt.MQTTv5,
        client_id="tools-cli",
    )
    client.username_pw_set(mqtt_username, mqtt_password)
    client.on_connect = on_connect
    client.on_publish = on_publish

    # Configure SSL/TLS
    client.tls_set()  # Default parameters for SSL. Adjust as necessary for your broker's configuration.

    # Connect to the MQTT Broker
    client.connect(mqtt_broker, mqtt_port, 60)

    # Start the network loop in a separate thread
    client.loop_start()

    try:
        while True:
            # Unpack the ZIP file and process each JSON file
            with tempfile.TemporaryDirectory() as temp_dir:
                with zipfile.ZipFile(input_file, "r") as zip_ref:
                    zip_ref.extractall(temp_dir)  # Extract files to the temporary directory
                    for root, dirs, files in os.walk(temp_dir):
                        sorted_files = sorted(files, key=lambda x: os.path.getmtime(os.path.join(root, x)))
                        for filename in sorted_files:
                            if filename.endswith(".json"):
                                file_path = os.path.join(root, filename)
                                with open(file_path, "r") as f:
                                    data = json.load(f)
                                    message = json.dumps(
                                        asdict(
                                            create_message(
                                                org_id,
                                                project_id,
                                                event_type,
                                                entity_type,
                                                entity_id,
                                                json.dumps(data),
                                            )
                                        )
                                    )
                                    # Uncomment the next line to publish
                                    client.publish(topic, payload=message, qos=1)
                                    logger.info(f"Published {filename}")

                                    sleep(0.1)
            if not infinite:
                break  # Exit the outer loop if infinite processing is not required
    finally:
        client.loop_stop()  # Stop the network loop
        client.disconnect()  # Disconnect from the MQTT broker

    client.loop_stop()
    client.disconnect()


simulate_game_match_cmd = simulate_game_match



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import typing as t
import uuid
from urllib.parse import urlparse

import click


def any_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: t.Any) -> t.Any:
    return value


def url_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: t.Any) -> t.Any:
    parsed = urlparse(value.strip())
    if not all([parsed.scheme, parsed.netloc]):
        click.BadParameter(f"Expected a valid URL, got '{value}' instead.")
    return value


def url_list_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: t.Any) -> t.List[str]:
    if isinstance(value, list):
        return value
    urls = value.split(",")
    for url in urls:
        parsed = urlparse(url.strip())
        if not all([parsed.scheme, parsed.netloc]):
            click.BadParameter(f"Expected a valid URL, got '{url}' instead.")
    return value


def username_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: t.Any) -> str:
    if not (6 <= len(value) < 200):
        click.BadParameter("Username must have at least 6 characters and less than 200 characters.")
    return value


def password_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: t.Any) -> str:
    if not (10 <= len(value) < 255):
        click.BadParameter("Password must have at least 10 characters and less than 255 characters.")
    return value


def uuid_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: t.Any) -> str:
    try:
        uuid.UUID(value)
    except ValueError:
        raise click.BadParameter(f"{value} is not a valid UUID.")
    return value


def kafka_brokers_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: str) -> str:
    """
    Validates that each broker in a comma-separated list is a valid host address.
    """
    brokers = value.split(",")
    for broker in brokers:
        parsed = urlparse("dummy-scheme://" + broker)
        if not parsed.hostname or not parsed.port:
            raise click.BadParameter(f"Each Kafka broker must be in the format 'host:port'. Error parsing '{broker}'.")
    return value


def zip_file_validator(ctx: click.Context, param: t.Optional[click.Parameter], value: str):
    """
    Validates that the provided file is a ZIP file based on its extension.
    """
    if not value.lower().endswith(".zip"):
        raise click.UsageError("The input file must be a ZIP file.")
    return value



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import click
from loguru import logger
from prettytable import PrettyTable

from stellanow_cli.core.validators import uuid_validator
from stellanow_cli.services.code_generator.code_generator import CodeGeneratorService, pass_code_generator_service


@click.command()
@click.option(
    "--project_id",
    required=True,
    prompt=True,
    callback=uuid_validator,
    help="UUID of the project associated with the organization saved in your configuration file.",
)
@pass_code_generator_service
def models(service: CodeGeneratorService, project_id: str, *args, **kwargs) -> None:
    """Fetches the latest models specifications from the API and output a list of the models into the terminal prompt."""
    workflow_client = service.create_workflow_client(project_id=project_id)
    _models = workflow_client.get_models()

    table = PrettyTable(["ModelID", "Model Name", "Created At", "Updated At"])

    for model in _models:
        table.add_row([model.id, model.name, model.createdAt, model.updatedAt])

    logger.info(table)

    for model in _models:
        logger.info(f"ID: {model.id}, Name: {model.name}")


models_cmd = models



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import functools
import typing as t
from functools import update_wrapper

import click
import typing_extensions as te
from click import get_current_context

from ..core.context import StellaNowContext
from ..services.service import StellaNowService

P = te.ParamSpec("P")
R = t.TypeVar("R")
T = t.TypeVar("T", bound=StellaNowService)
_AnyCallable = t.Callable[..., t.Any]


def option_with_config_lookup(
    param_name: str, service_class: t.Optional[t.Type[T]] = None, **option_kwargs
) -> t.Callable[[_AnyCallable], _AnyCallable]:
    def decorator(f):
        original_callback = option_kwargs.pop("callback", None)

        def callback(ctx, param, value):
            lookup_name = service_class if service_class else ctx.info_name
            stella_ctx: t.Optional[StellaNowContext] = ctx.find_object(StellaNowContext)

            if stella_ctx is None:
                raise click.ClickException(
                    f"{StellaNowContext.__class__.__name__} not found in the current Click context."
                )

            if value is None or ctx.get_parameter_source(param.name) == click.core.ParameterSource.DEFAULT:
                config_value = stella_ctx.get_config_value(lookup_name, param_name.replace("-", ""))
                value = config_value if config_value is not None else value

            if original_callback is not None:
                return original_callback(ctx, param, value)
            return value

        option_kwargs["callback"] = callback

        def default_value():
            ctx = click.get_current_context()
            lookup_service = service_class.service_name() if service_class else ctx.info_name
            stella_ctx: t.Optional[StellaNowContext] = ctx.find_object(StellaNowContext)

            if stella_ctx is None:
                raise click.ClickException(
                    f"{StellaNowContext.__class__.__name__} not found in the current Click context."
                )

            config_value = stella_ctx.get_config_value(lookup_service, param_name.replace("-", ""))
            if config_value is None:
                ctx.fail(f"Missing value for '{param_name}', provide configuration for Service: {lookup_service}")
            return config_value

        option_kwargs["default"] = functools.partial(default_value)
        option_kwargs["show_default"] = False

        return click.option(param_name, **option_kwargs)(f)

    return decorator


def prompt(param_name: str, **option_kwargs) -> t.Callable[[_AnyCallable], _AnyCallable]:
    """
    Creates a Click option decorator that automatically fetches its default value
    from the configuration, using a service name from ctx.obj.
    Additional keyword arguments are passed through to click.option().
    """

    def default_value():
        ctx = click.get_current_context()
        return ctx.obj.get_config_value(ctx.info_name, param_name.replace("-", ""))

    option_kwargs.setdefault("prompt", True)
    option_kwargs.setdefault("hidden", True)

    return click.option(param_name, default=default_value, **option_kwargs)


def make_stella_context_pass_decorator(
    object_type: t.Type[T],
) -> t.Callable[["t.Callable[te.Concatenate[T, P], R]"], "t.Callable[P, R]"]:
    def decorator(f: "t.Callable[te.Concatenate[T, P], R]") -> "t.Callable[P, R]":
        def new_func(*args: "P.args", **kwargs: "P.kwargs") -> "R":
            ctx = get_current_context()

            from ..core.context import StellaNowContext

            stella_ctx: t.Optional[StellaNowContext] = ctx.find_object(StellaNowContext)
            if stella_ctx is None:
                raise click.ClickException(
                    f"{StellaNowContext.__class__.__name__} not found in the current Click context."
                )

            obj: t.Optional[T]
            obj = stella_ctx.get_service(object_type)

            if obj is None:
                raise RuntimeError(
                    "Managed to invoke callback without a context"
                    f" object of type {object_type.__name__!r}"
                    " existing."
                )

            return ctx.invoke(f, obj, *args, **kwargs)

        return update_wrapper(new_func, f)

    return decorator  # type: ignore[return-value]



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from bs4 import BeautifulSoup
from keycloak import KeycloakError
from loguru import logger

from stellanow_cli.exceptions.cli_exceptions import StellaNowCLIException


class StellaAPIError(StellaNowCLIException):
    """Exception raised for errors in the Stella API."""

    def __init__(self, message, details=None):
        if details:
            logger.debug(f"Error: {details}")

        super().__init__(message, details)


class StellaAPIBadRequestError(StellaAPIError):
    """Exception raised for when a requested object does not exist in the Stella API."""

    def __init__(self, details):
        super().__init__("Bad Request", details)


class StellaAPIForbiddenError(StellaAPIError):
    """Exception raised for when trying to access the Stella API from blacklisted address."""

    def __init__(self, details):
        super().__init__("Forbidden", details)


class StellaAPINotFoundError(StellaAPIError):
    """Exception raised for when a requested object does not exist in the Stella API."""

    def __init__(self, details):
        super().__init__("Not Found", details)


class StellaAPIUnauthorisedError(StellaAPIError):
    """Exception raised for when request is not authorised to be performed by requesting entity in the Stella API."""

    def __init__(self, details):
        super().__init__("Unauthorised", details)


class StellaAPIWrongCredentialsError(StellaAPIError):
    """Exception raised for wrong credentials during auth in the Stella API."""

    def __init__(self):
        super().__init__("Unauthorized: Provided username or password is invalid.", {})


class StellaAPIInternalServerError(StellaAPIError):
    """Exception raised for when Internal Server Error."""

    def __init__(self, details):
        super().__init__("Internal Server Error.", details)


def parse_keycloak_error_response(error: KeycloakError):
    soup = BeautifulSoup(str(error), "html.parser")
    what_happened = soup.find_all("h2", text="What happened?")
    if what_happened:
        return what_happened[0].find_next("p").text
    h1_error = soup.find("h1")
    if h1_error:
        return h1_error.text
    return str(error)


class StellaNowApiKeycloakException(StellaAPIError):
    """Exception raised for Keycloak errors in the Stella Now API."""

    def __init__(self, message, details=None):
        if details:
            error_message = parse_keycloak_error_response(error=details)
            logger.debug(f"KeycloakError: {error_message}")

        super().__init__(message, details)


class StellaNowKeycloakCommunicationException(StellaNowApiKeycloakException):
    def __init__(self, details):
        super().__init__(
            f"An error occurred while trying to communicate with the authentication server",
            details,
        )



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

import inflection
from jinja2 import Environment, FileSystemLoader
from stellanow_api_internals.datatypes.workflow_mgmt import (
    StellaEventDetailed,
    StellaField,
    StellaFieldType,
    StellaModelDetailed,
)

from stellanow_cli.core.enums import FieldType, StellaLanguage


@dataclass
class CommentDefinition(ABC):
    start: str
    middle: str
    end: str
    single: str


@dataclass
class FieldDescriptor:
    original_name: str
    mapped_name: str
    type_str: str
    is_model: bool

    @classmethod
    def from_stella_field(cls, field, mapper, escape_reserved_words, model_details):
        field_type = mapper.map_field_type(field, model_details)
        is_model = field_type in mapper.model_types or field_type.endswith("Model")
        mapped_name = escape_reserved_words(field.name)
        return cls(field.name, mapped_name, field_type, is_model)


@dataclass
class EntityDescriptor:
    original_name: str
    mapped_name: str

    @classmethod
    def from_stella_entity(cls, entity, escape_reserved_words):
        return cls(entity.name, escape_reserved_words(entity.name))


@dataclass
class BaseClassSpec:
    id: str
    name: str
    json: str
    comment_def: CommentDefinition
    file_ext: str
    fields: List[FieldDescriptor]
    referenced_models: List[str]
    timestamp: str = field(default_factory=lambda: datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))


@dataclass
class ModelClassSpec(BaseClassSpec):
    pass


@dataclass
class MessageClassSpec(BaseClassSpec):
    entities: List[EntityDescriptor] = field(default_factory=list)


class LanguageConfig(ABC):
    @property
    @abstractmethod
    def language(self) -> StellaLanguage:
        pass

    @property
    @abstractmethod
    def type_mappings(self) -> Dict[FieldType, str]:
        pass

    @property
    @abstractmethod
    def reserved_words(self) -> set:
        pass

    @property
    @abstractmethod
    def comment_def(self) -> CommentDefinition:
        pass

    @property
    @abstractmethod
    def file_extension(self) -> str:
        pass

    def map_field_type(self, field: StellaField, model_details: Dict[str, Any]) -> str:
        if isinstance(field.fieldType, dict):
            field_type = StellaFieldType(**field.fieldType)
        else:
            field_type = field.fieldType

        value_type = FieldType(field_type.value)

        if value_type == FieldType.MODEL and field_type.modelRef:
            model_name = inflection.camelize(model_details[field_type.modelRef].name, uppercase_first_letter=True)
            return f"{model_name}Model"
        return self.type_mappings.get(value_type, FieldType.ANY)

    @property
    def model_types(self) -> List[str]:
        return ["Any", "object"]  # Default; override if needed

    def escape_reserved_words(self, word: str) -> str:
        return f"{word}_" if word in self.reserved_words else word

    def __post_init__(self):
        missing_types = set(FieldType) - set(self.type_mappings.keys())
        if missing_types:
            raise ValueError(f"LanguageConfig for {self.language} missing mappings for: {missing_types}")


class CodeGenerator(ABC):
    def __init__(self, language_config: LanguageConfig):
        self.config = language_config

        self.env = self.__load_templates()

        self.env.filters["camel_case"] = lambda s: inflection.camelize(s, uppercase_first_letter=False)
        self.env.filters["pascal_case"] = lambda s: inflection.camelize(s, uppercase_first_letter=True)
        self.env.filters["kebab_case"] = lambda s: inflection.dasherize
        self.env.filters["snake_case"] = inflection.underscore
        self.env.filters["get_model_file_name"] = self.get_file_name_for_model_name

    def __load_templates(self) -> Environment:
        current_file_path = Path(__file__).parent

        language_templates_path = current_file_path / f"../{self.get_language()}/templates"
        shared_templates_path = current_file_path / "templates"  # Shared templates directory

        return Environment(
            loader=FileSystemLoader([language_templates_path, shared_templates_path]),
            extensions=["jinja2.ext.do"],
        )

    def _prepare_fields(self, fields, model_details: Dict[str, Any]) -> List[FieldDescriptor]:
        return [
            FieldDescriptor.from_stella_field(f, self.config, self.config.escape_reserved_words, model_details)
            for f in fields
        ]

    def _prepare_entities(self, entities) -> List[EntityDescriptor]:
        return [EntityDescriptor.from_stella_entity(e, self.config.escape_reserved_words) for e in entities]

    def _collect_referenced_models(self, fields: List[FieldDescriptor]) -> List[str]:
        return [f.type_str[:-5] for f in fields if f.is_model and f.type_str.endswith("Model")]

    def generate_model_class(self, model, model_details: Dict[str, Any], **kwargs) -> str:
        template = self.env.get_template("model.jinja2")
        spec = self.get_model_data(model, model_details, **kwargs)
        return template.render(spec=spec)

    def generate_message_class(self, event, model_details: Dict[str, Any], **kwargs) -> str:
        template = self.env.get_template("message.jinja2")
        spec = self.get_message_data(event, model_details, **kwargs)
        return template.render(spec=spec)  # Allow spec to unpack as kwargs for flexibility

    @abstractmethod
    def get_model_data(
        self,
        model: StellaModelDetailed,
        model_details: Dict[str, StellaModelDetailed],
        **kwargs,
    ) -> ModelClassSpec:
        """Return a language-specific spec object for model class generation."""

    @abstractmethod
    def get_message_data(
        self,
        event: StellaEventDetailed,
        model_details: Dict[str, StellaModelDetailed],
        **kwargs,
    ) -> MessageClassSpec:
        """Return a language-specific spec object or dict for message class generation."""

    @abstractmethod
    def get_file_name_for_event_name(self, event_name: str) -> str:
        pass

    @abstractmethod
    def get_file_name_for_model_name(self, model_name: str) -> str:
        pass

    def get_language(self) -> str:
        return self.config.language.value



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from importlib.metadata import version

import click
from loguru import logger

from stellanow_cli.commands.code_generator.code_generator_service_group import code_generator_service
from stellanow_cli.commands.configure.configure_group import configure
from stellanow_cli.commands.data_dna_stream_tester.data_dna_stream_tester_group import data_dna_stream_tester
from stellanow_cli.core.context import DEFAULT, StellaNowContext, pass_stella_context

__version__ = version("stellanow-cli")


@click.group()
@click.version_option(version=__version__, message="%(version)s")
@click.option(
    "--profile",
    "-p",
    default=DEFAULT,
    help="The profile name for storing a particular set of configurations. If no profile is "
    "specified, the configurations will be stored under the 'DEFAULT' profile.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Enables verbose mode, which outputs more detailed logging messages.",
)
@pass_stella_context
def cli(stella_ctx: StellaNowContext, profile: str, verbose: bool) -> None:
    """Command-line interface for the StellaNow SDK code generation and comparison tool."""
    stella_ctx.profile = profile
    stella_ctx.verbose = verbose

    logger.info(f"Command executed in context of profile: {stella_ctx.profile} ")


cli.add_command(configure)
cli.add_command(code_generator_service)
cli.add_command(data_dna_stream_tester)



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import configparser
import os
import typing as t

from loguru import logger

STORAGE_FOLDER = ".stellanow"
CONFIG_FILE = "config.ini"


ProfileData = t.Dict[str, t.Any]


class ProfileManager:
    def __init__(self):
        self._loaded = False

        self.config: configparser.ConfigParser = configparser.ConfigParser(interpolation=None)

        self._profile: t.Optional[str] = None
        self.profile_data: ProfileData = {}  # This will store the loaded profile data

    def _load(self) -> None:
        home = os.path.expanduser("~")
        config_file = os.path.join(home, STORAGE_FOLDER, CONFIG_FILE)
        self.config.read(config_file)

        prefix = f"{self.profile}"
        for section in self.config.sections():
            if section.startswith(prefix):
                module_name = section[len(prefix) + 1 :]
                self.profile_data[module_name] = {k: v for k, v in self.config.items(section)}

    def save_service_config(self, service: str, service_data: ProfileData) -> None:
        home = os.path.expanduser("~")
        config_dir = os.path.join(home, STORAGE_FOLDER)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)  # Create the directory if it does not exist

        config_file = os.path.join(config_dir, CONFIG_FILE)
        self.config[f"{self.profile}:{service}"] = service_data

        # Save the updated configuration to the file
        with open(config_file, "w") as configfile:
            self.config.write(configfile)

        logger.success(f"Configuration for profile '{self.profile}' and service '{service}' saved successfully")

    @property
    def profile(self) -> t.Optional[str]:
        """Get the current profile."""
        return self._profile

    @profile.setter
    def profile(self, new_profile: str) -> None:
        """Set a new profile and reload data."""
        self._profile = new_profile
        self._load()  # Reload the data for the new profile

    def get_profile_data(self) -> ProfileData:
        return self.profile_data

    def get_service_data(self, service: str) -> ProfileData:
        return self.profile_data.get(service, {})



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import json
from dataclasses import dataclass, field
from typing import Dict

import inflection
from stellanow_api_internals.datatypes.workflow_mgmt import StellaEventDetailed, StellaModelDetailed

from stellanow_cli.code_generators.common.code_generator import (
    CodeGenerator,
    CommentDefinition,
    LanguageConfig,
    MessageClassSpec,
    ModelClassSpec,
)
from stellanow_cli.core.enums import FieldType, StellaLanguage
from stellanow_cli.exceptions.cli_exceptions import StellaNowCLINoEntityAssociatedWithEventException


@dataclass
class CSharpModelClassSpec(ModelClassSpec):
    namespace: str = field(default="StellaNowSDK")


@dataclass
class CSharpMessageClassSpec(MessageClassSpec):
    namespace: str = field(default="StellaNowSDK")


class CSharpLanguageConfig(LanguageConfig):
    @property
    def language(self) -> StellaLanguage:
        return StellaLanguage.CSHARP

    @property
    def type_mappings(self) -> Dict[FieldType, str]:
        return {
            FieldType.DECIMAL: "decimal",
            FieldType.INTEGER: "int",
            FieldType.BOOLEAN: "bool",
            FieldType.STRING: "string",
            FieldType.DATE: "DateOnly",
            FieldType.DATETIME: "DateTime",
            FieldType.MODEL: "object",
            FieldType.ANY: "object",
        }

    @property
    def reserved_words(self) -> set:
        return {
            "abstract",
            "as",
            "base",
            "bool",
            "break",
            "byte",
            "case",
            "catch",
            "char",
            "checked",
            "class",
            "const",
            "continue",
            "decimal",
            "default",
            "delegate",
            "do",
            "double",
            "else",
            "enum",
            "event",
            "explicit",
            "extern",
            "false",
            "finally",
            "fixed",
            "float",
            "for",
            "foreach",
            "goto",
            "if",
            "implicit",
            "in",
            "int",
            "interface",
            "internal",
            "is",
            "lock",
            "long",
            "namespace",
            "new",
            "null",
            "object",
            "operator",
            "out",
            "override",
            "params",
            "private",
            "protected",
            "public",
            "readonly",
            "ref",
            "return",
            "sbyte",
            "sealed",
            "short",
            "sizeof",
            "stackalloc",
            "static",
            "string",
            "struct",
            "switch",
            "this",
            "throw",
            "true",
            "try",
            "typeof",
            "uint",
            "ulong",
            "unchecked",
            "unsafe",
            "ushort",
            "using",
            "virtual",
            "void",
            "volatile",
            "while",
        }

    @property
    def comment_def(self) -> CommentDefinition:
        return CommentDefinition(start="/*", middle=" * ", end="*/", single="//")

    @property
    def file_extension(self) -> str:
        return "cs"


class CsharpCodeGenerator(CodeGenerator):
    def __init__(self):
        super().__init__(CSharpLanguageConfig())

    def get_model_data(
        self,
        model: StellaModelDetailed,
        model_details: Dict[str, StellaModelDetailed],
        **kwargs,
    ) -> CSharpModelClassSpec:
        fields = self._prepare_fields(model.fields.root, model_details)

        return CSharpModelClassSpec(
            id=model.id,
            name=model.name,
            json=json.dumps(model.model_dump(), indent=4),
            fields=fields,
            referenced_models=self._collect_referenced_models(fields),
            namespace=kwargs.get("namespace", "StellaNowSDK.Messages"),
            comment_def=self.config.comment_def,
            file_ext=self.config.file_extension,
        )

    def get_message_data(
        self,
        event: StellaEventDetailed,
        model_details: Dict[str, StellaModelDetailed],
        **kwargs,
    ) -> CSharpMessageClassSpec:
        fields = self._prepare_fields(event.fields, model_details)
        entities = self._prepare_entities(event.entities)
        if not entities:
            raise StellaNowCLINoEntityAssociatedWithEventException()

        return CSharpMessageClassSpec(
            id=event.id,
            name=event.name,
            json=json.dumps(event.model_dump(), indent=4),
            fields=fields,
            entities=entities,
            referenced_models=self._collect_referenced_models(fields),
            namespace=kwargs.get("namespace", "StellaNowSDK.Messages"),
            comment_def=self.config.comment_def,
            file_ext=self.config.file_extension,
        )

    def get_file_name_for_event_name(self, event_name: str) -> str:
        return f"{inflection.camelize(event_name, uppercase_first_letter=True)}Message"

    def get_file_name_for_model_name(self, model_name: str) -> str:
        return f"{inflection.camelize(model_name, uppercase_first_letter=True)}Model"



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import typing as t

import click

from ..services.service import StellaNowService
from .logger import setup_logging
from .profile_manager import ProfileData, ProfileManager

DEFAULT = "DEFAULT"

T = t.TypeVar("T", bound=StellaNowService)
ServiceType = t.Type[T]
ServiceDict = t.Dict[ServiceType, T]


class StellaNowContext:
    def __init__(self):
        self._profile_manager: ProfileManager = ProfileManager()
        self._profile: str = DEFAULT
        self._verbose: bool = False
        self._services: ServiceDict = {}

    @property
    def profile(self) -> str:
        return self._profile

    @profile.setter
    def profile(self, profile: str) -> None:
        self._profile = profile
        self._profile_manager.profile = profile

    def save_service_config(self, service_name: str, service_config: ProfileData) -> None:
        self._profile_manager.save_service_config(service_name, service_config)

    @property
    def verbose(self) -> bool:
        return self._verbose

    @verbose.setter
    def verbose(self, verbose: bool) -> None:
        self._verbose = verbose
        setup_logging(verbose)

    def service_profile_data(self, service_name: str) -> ProfileData:
        return self._profile_manager.get_service_data(service_name)

    def get_config_value(self, service_name: str, config_key: str, default=None) -> str:
        """Get a specific configuration value for a service."""
        service_data = self.service_profile_data(service_name)
        return service_data.get(config_key, default)

    def get_service(self, service_type: t.Type[T]) -> T:
        service = self._services.get(service_type)
        if service is None:
            raise RuntimeError(f"Service of type {service_type.__name__} not found")
        return service

    def add_service(self, service: StellaNowService):
        self._services[type(service)] = service


pass_stella_context = click.make_pass_decorator(StellaNowContext, ensure=True)



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import importlib
from typing import Any, List

import click
from loguru import logger

from stellanow_cli.core.enums import StellaLanguage
from stellanow_cli.core.utils.logger_utils import log_processed_and_skipped_result
from stellanow_cli.core.validators import uuid_validator
from stellanow_cli.exceptions.cli_exceptions import (
    StellaNowCLILanguageNotSupportedException,
    StellaNowCLINamespaceNotProvidedException,
)
from stellanow_cli.services.code_generator.code_generator import CodeGeneratorService, pass_code_generator_service
from stellanow_cli.services.code_generator.tools import (
    collect_all_model_references,
    collect_model_references,
    fetch_model_details,
    generate_event_files,
    generate_model_files,
)


@click.command()
@click.option("--namespace", "-n", default="", help="The namespace for the generated classes.")
@click.option(
    "--project_id",
    required=True,
    prompt=True,
    callback=uuid_validator,
    help="UUID of the project associated with the organization saved in your configuration file.",
)
@click.option(
    "--destination",
    "-d",
    default=".",
    help="The directory to save the generated classes.",
)
@click.option("--force", "-f", is_flag=True, help="Overwrite existing files.")
@click.option("--event_names", "-e", multiple=True, help="List of specific events to generate.")
@click.option(
    "--language",
    "-l",
    type=click.Choice([lang.value for lang in StellaLanguage], case_sensitive=False),
    default="csharp",
    help="The programming language for the generated classes.",
)
@pass_code_generator_service
def generate(
    service: CodeGeneratorService,
    project_id: str,
    destination: str,
    force: bool,
    event_names: List[str],
    language: str,
    **kwargs: dict[str, Any],
):
    """Fetches the latest event specifications from the API and generates corresponding class code in the desired programming language."""
    workflow_client = service.create_workflow_client(project_id=project_id)

    if language == "csharp" and not kwargs.get("namespace"):
        raise StellaNowCLINamespaceNotProvidedException

    logger.info("Generating...")

    generator_module = f"stellanow_cli.code_generators.{language}.code_generator"
    generator_class_name = f"{language.capitalize()}CodeGenerator"
    try:
        print(generator_module)
        print(generator_class_name)
        generator_class = getattr(importlib.import_module(generator_module), generator_class_name)
        generator = generator_class()  # Instantiate the generator
    except ImportError:
        raise StellaNowCLILanguageNotSupportedException(language)

    events = workflow_client.get_events(include_inactive=True)

    if event_names:
        events = [event for event in events if event.name in event_names]

    model_refs = collect_model_references(workflow_client=workflow_client, events=events)
    all_model_refs = collect_all_model_references(workflow_client=workflow_client, model_ids=model_refs)
    models_details = fetch_model_details(workflow_client=workflow_client, model_refs=all_model_refs)

    models_processed, models_skipped = generate_model_files(
        generator=generator,  # Pass the instance instead of the class
        models_details=models_details,
        destination=destination,
        force=force,
        **kwargs,
    )
    events_not_found, events_processed, events_skipped = generate_event_files(
        generator=generator,  # Pass the instance instead of the class
        events=events,
        event_names=event_names,
        workflow_client=workflow_client,
        models_details=models_details,
        destination=destination,
        force=force,
        **kwargs,
    )

    all_skipped_files = models_skipped + events_skipped
    log_processed_and_skipped_result(models_processed + events_processed, all_skipped_files, events_not_found)


generate_cmd = generate



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

from dataclasses import dataclass
from typing import Optional

from stellanow_api_internals.clients.workflow_manager_client import WorkflowManagerClient

from stellanow_cli.core.decorators import make_stella_context_pass_decorator
from stellanow_cli.services.service import StellaNowService, StellaNowServiceConfig

CODE_GENERATOR_SERVICE_NAME = "code-generator-service"
OIDC_CLIENT_ID = "tools-cli"


@dataclass
class CodeGeneratorServiceConfig(StellaNowServiceConfig):
    base_url: str
    username: str
    password: str
    organization_id: str


class CodeGeneratorService(StellaNowService):
    def __init__(self, config: CodeGeneratorServiceConfig) -> None:  # noqa
        self._config = config

    @classmethod
    def service_name(cls) -> str:
        return CODE_GENERATOR_SERVICE_NAME

    def create_workflow_client(self, project_id: Optional[str]) -> WorkflowManagerClient:
        return WorkflowManagerClient(
            base_url=self._config.base_url,
            username=self._config.username,
            password=self._config.password,
            organization_id=self._config.organization_id,
            project_id=project_id,
        )


pass_code_generator_service = make_stella_context_pass_decorator(CodeGeneratorService)



"""
Copyright (C) 2022-2025 Stella Technologies (UK) Limited.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import os
from typing import Any, Callable, Dict, List, Set, Tuple

from loguru import logger
from stellanow_api_internals.clients.workflow_manager_client import WorkflowManagerClient
from stellanow_api_internals.datatypes.workflow_mgmt import (
    StellaEvent,
    StellaEventDetailed,
    StellaFieldType,
    StellaModelDetailed,
    StellaModelFieldType,
)

from stellanow_cli.code_generators.common.code_generator import CodeGenerator, LanguageConfig
from stellanow_cli.core.enums import StellaDataStructure
from stellanow_cli.core.helpers import ProcessedFile, SkippedFile, ensure_destination_exists
from stellanow_cli.exceptions.cli_exceptions import StellaNowCLIException


def collect_model_references(workflow_client: WorkflowManagerClient, events: List[StellaEvent]) -> Set[str]:
    model_refs = set()
    for event in events:
        detailed_event = workflow_client.get_event_details(event.id)
        for field in detailed_event.fields:
            if isinstance(field.fieldType, dict):
                field_type = StellaFieldType(**field.fieldType)
            else:
                field_type = field.fieldType

            if field_type.value == StellaDataStructure.MODEL:
                model_refs.add(field_type.modelRef)
    return model_refs


def collect_all_model_references(workflow_client: WorkflowManagerClient, model_ids: Set[str]) -> Set[str]:
    all_model_refs = set(model_ids)
    checked_model_refs = set()

    while model_ids:
        current_id = model_ids.pop()
        if current_id in checked_model_refs:
            continue

        model_details = workflow_client.get_model_details(current_id)
        for field in model_details.fields.root:
            if isinstance(field.fieldType, Dict):
                field_type = StellaModelFieldType(**field.fieldType)
            else:
                field_type = field.fieldType
            if field_type.value == StellaDataStructure.MODEL and field_type.modelRef:
                if field_type.modelRef not in all_model_refs:
                    model_ids.add(field_type.modelRef)
                    all_model_refs.add(field_type.modelRef)

        checked_model_refs.add(current_id)

    return all_model_refs


def fetch_model_details(workflow_client: WorkflowManagerClient, model_refs: set) -> Dict[str, StellaModelDetailed]:
    models_details: Dict[str, StellaModelDetailed] = {}
    for model_id in model_refs:
        models_details[model_id] = workflow_client.get_model_details(model_id)
    return models_details


def _generate_files(
    items: List[Any],
    generate_method: Callable[[Any, Dict[str, StellaModelDetailed], Any], str],
    get_file_name_method: Callable[[str], str],
    lang_conf: LanguageConfig,
    models_details: Dict[str, StellaModelDetailed],
    destination: str,
    force: bool,
    is_model: bool = False,
    **kwargs,
) -> Tuple[List[ProcessedFile], List[SkippedFile]]:
    processed_files = []
    skipped_files = []

    for item in items:
        logger.info(f"Generating class for item: {item.name}")

        if is_model:
            file_path = os.path.join(
                destination, "Models", f"{get_file_name_method(item.name)}.{lang_conf.file_extension}"
            )
        else:
            file_path = os.path.join(destination, f"{get_file_name_method(item.name)}.{lang_conf.file_extension}")

        if not force and os.path.exists(file_path):
            logger.info("Skipped ...")
            skipped_files.append(SkippedFile(item.name, "File Already Exist"))
            continue

        try:
            code = generate_method(item, models_details, **kwargs)
            ensure_destination_exists(os.path.dirname(file_path))
            with open(file_path, "w") as file:
                file.write(code)
            processed_files.append(ProcessedFile(item.name))
        except StellaNowCLIException as e:
            logger.info("Skipped ...")
            skipped_files.append(SkippedFile(item.name, e.message))

    return processed_files, skipped_files


def generate_model_files(
    generator: CodeGenerator,
    models_details: Dict[str, StellaModelDetailed],
    destination: str,
    force: bool,
    **kwargs,
) -> Tuple[List[ProcessedFile], List[SkippedFile]]:
    return _generate_files(
        items=list(models_details.values()),
        generate_method=generator.generate_model_class,
        get_file_name_method=generator.get_file_name_for_model_name,
        lang_conf=generator.config,
        models_details=models_details,
        destination=destination,
        force=force,
        is_model=True,
        **kwargs,
    )


def generate_event_files(
    generator: CodeGenerator,
    events: List[StellaEvent],
    event_names: List[str],
    workflow_client: WorkflowManagerClient,
    models_details: Dict[str, StellaModelDetailed],
    destination: str,
    force: bool,
    **kwargs,
) -> Tuple[Set[str], List[ProcessedFile], List[SkippedFile]]:
    events_not_found = set(event_names) if event_names else set()
    events_to_generate: List[StellaEventDetailed] = []

    for event in events:
        if event_names and event.name not in event_names:
            continue
        event_detailed = workflow_client.get_event_details(event.id)
        events_not_found.discard(event.name)
        events_to_generate.append(event_detailed)

    processed_files, skipped_files = _generate_files(
        items=events_to_generate,
        generate_method=generator.generate_message_class,
        get_file_name_method=generator.get_file_name_for_event_name,
        lang_conf=generator.config,
        models_details=models_details,
        destination=destination,
        force=force,
        **kwargs,
    )

    return events_not_found, processed_files, skipped_files
